from numpy import int_, str_, float_


Hexel = int
Sensor = str
Latency = float

# Set consistent dtypes for use in arrays
HexelDType = int_
SensorDType = str_
LatencyDType = float_
FunctionNameDType = str_
